@@ -1,21 +1,27 @@
 """Tests related to uaclient.entitlement.base module."""
 
+import contextlib
 import copy
+import io
 import logging
-import mock
+from functools import partial
 from types import MappingProxyType
 
+import mock
 import pytest
 
-from uaclient import apt
-from uaclient import exceptions
+from uaclient import apt, exceptions, messages, status
 from uaclient.entitlements.livepatch import (
+    LIVEPATCH_CMD,
     LivepatchEntitlement,
+    configure_livepatch_proxy,
+    get_config_option_value,
     process_config_directives,
+    unconfigure_livepatch_proxy,
 )
 from uaclient.entitlements.tests.conftest import machine_token
-from uaclient import status
-from uaclient.status import ContractStatus
+from uaclient.snap import SNAP_CMD
+from uaclient.status import ApplicationStatus, ContractStatus
 
 PLATFORM_INFO_SUPPORTED = MappingProxyType(
     {
@@ -27,33 +33,203 @@ PLATFORM_INFO_SUPPORTED = MappingProxyType(
 )
 
 M_PATH = "uaclient.entitlements.livepatch."  # mock path
+M_LIVEPATCH_STATUS = M_PATH + "LivepatchEntitlement.application_status"
+DISABLED_APP_STATUS = (status.ApplicationStatus.DISABLED, "")
+
 M_BASE_PATH = "uaclient.entitlements.base.UAEntitlement."
 
+DEFAULT_AFFORDANCES = {
+    "architectures": ["x86_64"],
+    "minKernelVersion": "4.4",
+    "kernelFlavors": ["generic", "lowlatency"],
+    "tier": "stable",
+}
+
 
 @pytest.fixture
-def entitlement(entitlement_factory):
-    affordances = {
-        "architectures": ["x86_64"],
-        "minKernelVersion": "4.4",
-        "kernelFlavors": ["generic", "lowlatency"],
-        "tier": "stable",
-    }
+def livepatch_entitlement_factory(entitlement_factory):
     directives = {"caCerts": "", "remoteServer": "https://alt.livepatch.com"}
-    return entitlement_factory(
-        LivepatchEntitlement, affordances=affordances, directives=directives
+    return partial(
+        entitlement_factory,
+        LivepatchEntitlement,
+        affordances=DEFAULT_AFFORDANCES,
+        directives=directives,
     )
 
 
+@pytest.fixture
+def entitlement(livepatch_entitlement_factory):
+    return livepatch_entitlement_factory()
+
+
+class TestConfigureLivepatchProxy:
+    @pytest.mark.parametrize(
+        "http_proxy,https_proxy,retry_sleeps",
+        (
+            ("http_proxy", "https_proxy", [1, 2]),
+            ("http_proxy", "", None),
+            ("", "https_proxy", [1, 2]),
+            ("http_proxy", None, [1, 2]),
+            (None, "https_proxy", None),
+            (None, None, [1, 2]),
+        ),
+    )
+    @mock.patch("uaclient.util.subp")
+    def test_configure_livepatch_proxy(
+        self, m_subp, http_proxy, https_proxy, retry_sleeps, capsys
+    ):
+        configure_livepatch_proxy(http_proxy, https_proxy, retry_sleeps)
+        expected_calls = []
+        if http_proxy:
+            expected_calls.append(
+                mock.call(
+                    [
+                        LIVEPATCH_CMD,
+                        "config",
+                        "http-proxy={}".format(http_proxy),
+                    ],
+                    retry_sleeps=retry_sleeps,
+                )
+            )
+
+        if https_proxy:
+            expected_calls.append(
+                mock.call(
+                    [
+                        LIVEPATCH_CMD,
+                        "config",
+                        "https-proxy={}".format(https_proxy),
+                    ],
+                    retry_sleeps=retry_sleeps,
+                )
+            )
+
+        assert m_subp.call_args_list == expected_calls
+
+        out, _ = capsys.readouterr()
+        if http_proxy or https_proxy:
+            assert out.strip() == messages.SETTING_SERVICE_PROXY.format(
+                service=LivepatchEntitlement.title
+            )
+
+    @pytest.mark.parametrize(
+        "key, subp_return_value, expected_ret",
+        [
+            ("http-proxy", ("nonsense", ""), None),
+            ("http-proxy", ("", "nonsense"), None),
+            (
+                "http-proxy",
+                (
+                    """\
+http-proxy: ""
+https-proxy: ""
+no-proxy: ""
+remote-server: https://livepatch.canonical.com
+ca-certs: ""
+check-interval: 60  # minutes""",
+                    "",
+                ),
+                None,
+            ),
+            (
+                "http-proxy",
+                (
+                    """\
+http-proxy: one
+https-proxy: two
+no-proxy: ""
+remote-server: https://livepatch.canonical.com
+ca-certs: ""
+check-interval: 60  # minutes""",
+                    "",
+                ),
+                "one",
+            ),
+            (
+                "https-proxy",
+                (
+                    """\
+http-proxy: one
+https-proxy: two
+no-proxy: ""
+remote-server: https://livepatch.canonical.com
+ca-certs: ""
+check-interval: 60  # minutes""",
+                    "",
+                ),
+                "two",
+            ),
+            (
+                "nonexistentkey",
+                (
+                    """\
+http-proxy: one
+https-proxy: two
+no-proxy: ""
+remote-server: https://livepatch.canonical.com
+ca-certs: ""
+check-interval: 60  # minutes""",
+                    "",
+                ),
+                None,
+            ),
+        ],
+    )
+    @mock.patch("uaclient.util.subp")
+    def test_get_config_option_value(
+        self, m_util_subp, key, subp_return_value, expected_ret
+    ):
+        m_util_subp.return_value = subp_return_value
+        ret = get_config_option_value(key)
+        assert ret == expected_ret
+        assert [
+            mock.call([LIVEPATCH_CMD, "config"])
+        ] == m_util_subp.call_args_list
+
+
+class TestUnconfigureLivepatchProxy:
+    @pytest.mark.parametrize(
+        "livepatch_installed, protocol_type, retry_sleeps",
+        (
+            (True, "http", None),
+            (True, "https", [1]),
+            (True, "http", []),
+            (False, "http", None),
+        ),
+    )
+    @mock.patch("uaclient.util.which")
+    @mock.patch("uaclient.util.subp")
+    def test_unconfigure_livepatch_proxy(
+        self, subp, which, livepatch_installed, protocol_type, retry_sleeps
+    ):
+        if livepatch_installed:
+            which.return_value = LIVEPATCH_CMD
+        else:
+            which.return_value = None
+        kwargs = {"protocol_type": protocol_type}
+        if retry_sleeps is not None:
+            kwargs["retry_sleeps"] = retry_sleeps
+        assert None is unconfigure_livepatch_proxy(**kwargs)
+        if livepatch_installed:
+            expected_calls = [
+                mock.call(
+                    [LIVEPATCH_CMD, "config", protocol_type + "-proxy="],
+                    retry_sleeps=retry_sleeps,
+                )
+            ]
+        else:
+            expected_calls = []
+        assert expected_calls == subp.call_args_list
+
+
 class TestLivepatchContractStatus:
     def test_contract_status_entitled(self, entitlement):
         """The contract_status returns ENTITLED when entitled is True."""
         assert ContractStatus.ENTITLED == entitlement.contract_status()
 
-    def test_contract_status_unentitled(self, entitlement):
+    def test_contract_status_unentitled(self, livepatch_entitlement_factory):
         """The contract_status returns NONE when entitled is False."""
-        entitlement.cfg.write_cache(
-            "machine-access-livepatch", {"entitlement": {"entitled": False}}
-        )
+        entitlement = livepatch_entitlement_factory(entitled=False)
         assert ContractStatus.UNENTITLED == entitlement.contract_status()
 
 
@@ -62,16 +238,13 @@ class TestLivepatchUserFacingStatus:
         "uaclient.entitlements.livepatch.util.is_container", return_value=False
     )
     def test_user_facing_status_inapplicable_on_inapplicable_status(
-        self, _m_is_container, entitlement
+        self, _m_is_container, livepatch_entitlement_factory
     ):
         """The user-facing details INAPPLICABLE applicability_status"""
-        livepatch_bionic = entitlement.cfg.read_cache(
-            "machine-access-livepatch"
-        )
-        livepatch_bionic["entitlement"]["affordances"]["series"] = ["bionic"]
-        entitlement.cfg.write_cache(
-            "machine-access-livepatch", livepatch_bionic
-        )
+        affordances = copy.deepcopy(DEFAULT_AFFORDANCES)
+        affordances["series"] = ["bionic"]
+
+        entitlement = livepatch_entitlement_factory(affordances=affordances)
 
         with mock.patch("uaclient.util.get_platform_info") as m_platform_info:
             m_platform_info.return_value = PLATFORM_INFO_SUPPORTED
@@ -81,7 +254,7 @@ class TestLivepatchUserFacingStatus:
             "Livepatch is not available for Ubuntu 16.04 LTS"
             " (Xenial Xerus)."
         )
-        assert expected_details == details
+        assert expected_details == details.msg
 
     def test_user_facing_status_unavailable_on_unentitled(self, entitlement):
         """Status UNAVAILABLE on absent entitlement contract status."""
@@ -96,7 +269,7 @@ class TestLivepatchUserFacingStatus:
             m_platform_info.return_value = PLATFORM_INFO_SUPPORTED
             uf_status, details = entitlement.user_facing_status()
         assert uf_status == status.UserFacingStatus.UNAVAILABLE
-        assert "Livepatch is not entitled" == details
+        assert "Livepatch is not entitled" == details.msg
 
 
 class TestLivepatchProcessConfigDirectives:
@@ -114,7 +287,7 @@ class TestLivepatchProcessConfigDirectives:
             process_config_directives(cfg)
         expected_subp = mock.call(
             [
-                "/snap/bin/canonical-livepatch",
+                LIVEPATCH_CMD,
                 "config",
                 livepatch_param_tmpl.format(directive_value),
             ],
@@ -137,16 +310,10 @@ class TestLivepatchProcessConfigDirectives:
             process_config_directives(cfg)
         expected_calls = [
             mock.call(
-                ["/snap/bin/canonical-livepatch", "config", "ca-certs=value2"],
-                capture=True,
+                [LIVEPATCH_CMD, "config", "ca-certs=value2"], capture=True
             ),
             mock.call(
-                [
-                    "/snap/bin/canonical-livepatch",
-                    "config",
-                    "remote-server=value1",
-                ],
-                capture=True,
+                [LIVEPATCH_CMD, "config", "remote-server=value1"], capture=True
             ),
         ]
         assert expected_calls == m_subp.call_args_list
@@ -160,6 +327,11 @@ class TestLivepatchProcessConfigDirectives:
         assert 0 == m_subp.call_count
 
 
+@mock.patch(
+    "uaclient.entitlements.fips.FIPSEntitlement.application_status",
+    return_value=DISABLED_APP_STATUS,
+)
+@mock.patch(M_LIVEPATCH_STATUS, return_value=DISABLED_APP_STATUS)
 @mock.patch(
     "uaclient.entitlements.livepatch.util.is_container", return_value=False
 )
@@ -169,7 +341,13 @@ class TestLivepatchEntitlementCanEnable:
         ("4.4.0-00-generic", "5.0.0-00-generic", "4.19.0-00-generic"),
     )
     def test_can_enable_true_on_entitlement_inactive(
-        self, _m_is_container, supported_kernel_ver, capsys, entitlement
+        self,
+        _m_is_container,
+        _m_livepatch_status,
+        _m_fips_status,
+        supported_kernel_ver,
+        capsys,
+        entitlement,
     ):
         """When entitlement is INACTIVE, can_enable returns True."""
         supported_kernel = copy.deepcopy(dict(PLATFORM_INFO_SUPPORTED))
@@ -178,12 +356,12 @@ class TestLivepatchEntitlementCanEnable:
             with mock.patch("uaclient.util.is_container") as m_container:
                 m_platform.return_value = supported_kernel
                 m_container.return_value = False
-                assert entitlement.can_enable()
+                assert (True, None) == entitlement.can_enable()
         assert ("", "") == capsys.readouterr()
         assert [mock.call()] == m_container.call_args_list
 
     def test_can_enable_false_on_unsupported_kernel_min_version(
-        self, _m_is_container, capsys, entitlement
+        self, _m_is_container, _m_livepatch_status, _m_fips_status, entitlement
     ):
         """"False when on a kernel less or equal to minKernelVersion."""
         unsupported_min_kernel = copy.deepcopy(dict(PLATFORM_INFO_SUPPORTED))
@@ -191,15 +369,17 @@ class TestLivepatchEntitlementCanEnable:
         with mock.patch("uaclient.util.get_platform_info") as m_platform:
             m_platform.return_value = unsupported_min_kernel
             entitlement = LivepatchEntitlement(entitlement.cfg)
-            assert not entitlement.can_enable()
-        msg = (
-            "Livepatch is not available for kernel 4.2.9-00-generic.\n"
-            "Minimum kernel version required: 4.4\n"
-        )
-        assert (msg, "") == capsys.readouterr()
+            result, reason = entitlement.can_enable()
+            assert False is result
+            assert status.CanEnableFailureReason.INAPPLICABLE == reason.reason
+            msg = (
+                "Livepatch is not available for kernel 4.2.9-00-generic.\n"
+                "Minimum kernel version required: 4.4."
+            )
+            assert msg == reason.message.msg
 
     def test_can_enable_false_on_unsupported_kernel_flavor(
-        self, _m_is_container, capsys, entitlement
+        self, _m_is_container, _m_livepatch_status, _m_fips_status, entitlement
     ):
         """"When on an unsupported kernel, can_enable returns False."""
         unsupported_kernel = copy.deepcopy(dict(PLATFORM_INFO_SUPPORTED))
@@ -207,12 +387,14 @@ class TestLivepatchEntitlementCanEnable:
         with mock.patch("uaclient.util.get_platform_info") as m_platform:
             m_platform.return_value = unsupported_kernel
             entitlement = LivepatchEntitlement(entitlement.cfg)
-            assert not entitlement.can_enable()
-        msg = (
-            "Livepatch is not available for kernel 4.4.0-140-notgeneric.\n"
-            "Supported flavors are: generic, lowlatency\n"
-        )
-        assert (msg, "") == capsys.readouterr()
+            result, reason = entitlement.can_enable()
+            assert False is result
+            assert status.CanEnableFailureReason.INAPPLICABLE == reason.reason
+            msg = (
+                "Livepatch is not available for kernel 4.4.0-140-notgeneric.\n"
+                "Supported flavors are: generic, lowlatency."
+            )
+            assert msg == reason.message.msg
 
     @pytest.mark.parametrize(
         "kernel_version,meets_min_version",
@@ -227,9 +409,10 @@ class TestLivepatchEntitlementCanEnable:
     def test_can_enable_false_on_unsupported_min_kernel_version(
         self,
         _m_is_container,
+        _m_livepatch_status,
+        _m_fips_status,
         kernel_version,
         meets_min_version,
-        capsys,
         entitlement,
     ):
         """"When on an unsupported kernel version, can_enable returns False."""
@@ -239,35 +422,40 @@ class TestLivepatchEntitlementCanEnable:
             m_platform.return_value = unsupported_kernel
             entitlement = LivepatchEntitlement(entitlement.cfg)
             if meets_min_version:
-                assert entitlement.can_enable()
+                assert (True, None) == entitlement.can_enable()
             else:
-                assert not entitlement.can_enable()
-        if meets_min_version:
-            msg = ""
-        else:
-            msg = (
-                "Livepatch is not available for kernel {}.\n"
-                "Minimum kernel version required: 4.4\n".format(kernel_version)
-            )
-        assert (msg, "") == capsys.readouterr()
+                result, reason = entitlement.can_enable()
+                assert False is result
+                assert (
+                    status.CanEnableFailureReason.INAPPLICABLE == reason.reason
+                )
+                msg = (
+                    "Livepatch is not available for kernel {}.\n"
+                    "Minimum kernel version required: 4.4.".format(
+                        kernel_version
+                    )
+                )
+                assert msg == reason.message.msg
 
     def test_can_enable_false_on_unsupported_architecture(
-        self, _m_is_container, capsys, entitlement
+        self, _m_is_container, _m_livepatch_status, _m_fips_status, entitlement
     ):
         """"When on an unsupported architecture, can_enable returns False."""
         unsupported_kernel = copy.deepcopy(dict(PLATFORM_INFO_SUPPORTED))
         unsupported_kernel["arch"] = "ppc64le"
         with mock.patch("uaclient.util.get_platform_info") as m_platform:
             m_platform.return_value = unsupported_kernel
-            assert not entitlement.can_enable()
-        msg = (
-            "Livepatch is not available for platform ppc64le.\n"
-            "Supported platforms are: x86_64\n"
-        )
-        assert (msg, "") == capsys.readouterr()
+            result, reason = entitlement.can_enable()
+            assert False is result
+            assert status.CanEnableFailureReason.INAPPLICABLE == reason.reason
+            msg = (
+                "Livepatch is not available for platform ppc64le.\n"
+                "Supported platforms are: x86_64."
+            )
+            assert msg == reason.message.msg
 
     def test_can_enable_false_on_containers(
-        self, m_is_container, capsys, entitlement
+        self, m_is_container, _m_livepatch_status, _m_fips_status, entitlement
     ):
         """When is_container is True, can_enable returns False."""
         unsupported_min_kernel = copy.deepcopy(dict(PLATFORM_INFO_SUPPORTED))
@@ -276,9 +464,11 @@ class TestLivepatchEntitlementCanEnable:
             m_platform.return_value = unsupported_min_kernel
             m_is_container.return_value = True
             entitlement = LivepatchEntitlement(entitlement.cfg)
-            assert not entitlement.can_enable()
-        msg = "Cannot install Livepatch on a container\n"
-        assert (msg, "") == capsys.readouterr()
+            result, reason = entitlement.can_enable()
+            assert False is result
+            assert status.CanEnableFailureReason.INAPPLICABLE == reason.reason
+            msg = "Cannot install Livepatch on a container."
+            assert msg == reason.message.msg
 
 
 class TestLivepatchProcessContractDeltas:
@@ -293,14 +483,14 @@ class TestLivepatchProcessContractDeltas:
     @mock.patch(M_PATH + "LivepatchEntitlement.setup_livepatch_config")
     @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
     @mock.patch(M_PATH + "LivepatchEntitlement.applicability_status")
-    def test_true_on_inactive_livepatch_service(
+    def test_false_on_inactive_livepatch_service(
         self,
         m_applicability_status,
         m_application_status,
         m_setup_livepatch_config,
         entitlement,
     ):
-        """When livepatch is INACTIVE return True and do no setup."""
+        """When livepatch is INACTIVE return False and do no setup."""
         m_applicability_status.return_value = (
             status.ApplicabilityStatus.APPLICABLE,
             "",
@@ -310,7 +500,7 @@ class TestLivepatchProcessContractDeltas:
             "",
         )
         deltas = {"entitlement": {"directives": {"caCerts": "new"}}}
-        assert entitlement.process_contract_deltas({}, deltas, False)
+        assert not entitlement.process_contract_deltas({}, deltas, False)
         assert [] == m_setup_livepatch_config.call_args_list
 
     @pytest.mark.parametrize(
@@ -382,20 +572,23 @@ class TestLivepatchProcessContractDeltas:
         assert setup_calls == m_setup_livepatch_config.call_args_list
 
 
+@mock.patch("uaclient.util.validate_proxy", side_effect=lambda x, y, z: y)
+@mock.patch("uaclient.snap.configure_snap_proxy")
+@mock.patch("uaclient.entitlements.livepatch.configure_livepatch_proxy")
 class TestLivepatchEntitlementEnable:
 
-    mocks_apt_update = [
-        mock.call(["apt-get", "update"], status.MESSAGE_APT_UPDATE_FAILED)
-    ]
+    mocks_apt_update = [mock.call()]
     mocks_snapd_install = [
         mock.call(
             ["apt-get", "install", "--assume-yes", "snapd"],
             capture=True,
             retry_sleeps=apt.APT_RETRIES,
-        ),
+        )
+    ]
+    mocks_snap_wait_seed = [
         mock.call(
             ["/usr/bin/snap", "wait", "system", "seed.loaded"], capture=True
-        ),
+        )
     ]
     mocks_livepatch_install = [
         mock.call(
@@ -404,60 +597,45 @@ class TestLivepatchEntitlementEnable:
             retry_sleeps=[0.5, 1, 5],
         )
     ]
-    mocks_install = mocks_snapd_install + mocks_livepatch_install
+    mocks_install = (
+        mocks_snapd_install + mocks_snap_wait_seed + mocks_livepatch_install
+    )
     mocks_config = [
         mock.call(
             [
-                "/snap/bin/canonical-livepatch",
+                LIVEPATCH_CMD,
                 "config",
                 "remote-server=https://alt.livepatch.com",
             ],
             capture=True,
         ),
-        mock.call(["/snap/bin/canonical-livepatch", "disable"]),
-        mock.call(
-            ["/snap/bin/canonical-livepatch", "enable", "TOKEN"], capture=True
-        ),
+        mock.call([LIVEPATCH_CMD, "disable"]),
+        mock.call([LIVEPATCH_CMD, "enable", "livepatch-token"], capture=True),
     ]
 
-    @mock.patch(M_PATH + "LivepatchEntitlement.can_enable", return_value=False)
-    def test_enable_false_when_can_enable_false(
-        self, m_can_enable, caplog_text, capsys, entitlement
-    ):
-        """When can_enable returns False enable returns False."""
-        assert not entitlement.enable()
-        assert "" == caplog_text()
-        assert ("", "") == capsys.readouterr()  # No additional prints
-        assert [mock.call(silent=mock.ANY)] == m_can_enable.call_args_list
-
-    @pytest.mark.parametrize("silent_if_inapplicable", (True, False, None))
-    @mock.patch(M_PATH + "LivepatchEntitlement.can_enable", return_value=False)
-    def test_enable_passes_silent_if_inapplicable_through(
-        self, m_can_enable, caplog_text, entitlement, silent_if_inapplicable
-    ):
-        """When can_enable returns False enable returns False."""
-        kwargs = {}
-        if silent_if_inapplicable is not None:
-            kwargs["silent_if_inapplicable"] = silent_if_inapplicable
-        entitlement.enable(**kwargs)
-
-        expected_call = mock.call(silent=bool(silent_if_inapplicable))
-        assert [expected_call] == m_can_enable.call_args_list
-
     @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
     @pytest.mark.parametrize("apt_update_success", (True, False))
+    @mock.patch("uaclient.util.get_platform_info")
     @mock.patch("uaclient.util.subp")
-    @mock.patch("uaclient.apt.run_apt_command")
+    @mock.patch("uaclient.apt.run_apt_install_command")
+    @mock.patch("uaclient.apt.run_apt_update_command")
     @mock.patch("uaclient.util.which", return_value=False)
     @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
-    @mock.patch(M_PATH + "LivepatchEntitlement.can_enable", return_value=True)
+    @mock.patch(
+        M_PATH + "LivepatchEntitlement.can_enable", return_value=(True, None)
+    )
     def test_enable_installs_snapd_and_livepatch_snap_when_absent(
         self,
         m_can_enable,
         m_app_status,
         m_which,
-        m_run_apt,
+        m_run_apt_update,
+        m_run_apt_install,
         m_subp,
+        _m_get_platform_info,
+        m_livepatch_proxy,
+        m_snap_proxy,
+        m_validate_proxy,
         capsys,
         caplog_text,
         entitlement,
@@ -467,16 +645,16 @@ class TestLivepatchEntitlementEnable:
         application_status = status.ApplicationStatus.ENABLED
         m_app_status.return_value = application_status, "enabled"
 
-        def fake_run_apt(cmd, message):
+        def fake_run_apt_update():
             if apt_update_success:
                 return
             raise exceptions.UserFacingError("Apt go BOOM")
 
-        m_run_apt.side_effect = fake_run_apt
+        m_run_apt_update.side_effect = fake_run_apt_update
 
         assert entitlement.enable()
         assert self.mocks_install + self.mocks_config in m_subp.call_args_list
-        assert self.mocks_apt_update == m_run_apt.call_args_list
+        assert self.mocks_apt_update == m_run_apt_update.call_args_list
         msg = (
             "Installing snapd\n"
             "Updating package lists\n"
@@ -492,27 +670,42 @@ class TestLivepatchEntitlementEnable:
             assert expected_log not in caplog_text()
         else:
             assert expected_log in caplog_text()
-        expected_calls = [
-            mock.call("/snap/bin/canonical-livepatch"),
-            mock.call("/usr/bin/snap"),
-        ]
+        expected_calls = [mock.call("/usr/bin/snap"), mock.call(LIVEPATCH_CMD)]
         assert expected_calls == m_which.call_args_list
+        assert m_validate_proxy.call_count == 2
+        assert m_snap_proxy.call_count == 1
+        assert m_livepatch_proxy.call_count == 1
 
+    @mock.patch("uaclient.util.get_platform_info")
     @mock.patch("uaclient.util.subp", return_value=("snapd", ""))
     @mock.patch(
         "uaclient.util.which", side_effect=lambda cmd: cmd == "/usr/bin/snap"
     )
     @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
-    @mock.patch(M_PATH + "LivepatchEntitlement.can_enable", return_value=True)
+    @mock.patch(
+        M_PATH + "LivepatchEntitlement.can_enable", return_value=(True, None)
+    )
     def test_enable_installs_only_livepatch_snap_when_absent_but_snapd_present(
-        self, m_can_enable, m_app_status, m_which, m_subp, capsys, entitlement
+        self,
+        m_can_enable,
+        m_app_status,
+        m_which,
+        m_subp,
+        _m_get_platform_info,
+        m_livepatch_proxy,
+        m_snap_proxy,
+        m_validate_proxy,
+        capsys,
+        entitlement,
     ):
         """Install canonical-livepatch snap when not present on the system."""
         application_status = status.ApplicationStatus.ENABLED
         m_app_status.return_value = application_status, "enabled"
         assert entitlement.enable()
         assert (
-            self.mocks_livepatch_install + self.mocks_config
+            self.mocks_snap_wait_seed
+            + self.mocks_livepatch_install
+            + self.mocks_config
             in m_subp.call_args_list
         )
         msg = (
@@ -520,20 +713,31 @@ class TestLivepatchEntitlementEnable:
             "Canonical livepatch enabled.\n"
         )
         assert (msg, "") == capsys.readouterr()
-        expected_calls = [
-            mock.call("/snap/bin/canonical-livepatch"),
-            mock.call("/usr/bin/snap"),
-        ]
+        expected_calls = [mock.call("/usr/bin/snap"), mock.call(LIVEPATCH_CMD)]
         assert expected_calls == m_which.call_args_list
+        assert m_validate_proxy.call_count == 2
+        assert m_snap_proxy.call_count == 1
+        assert m_livepatch_proxy.call_count == 1
 
     @mock.patch("uaclient.util.subp")
     @mock.patch(
         "uaclient.util.which", side_effect=lambda cmd: cmd == "/usr/bin/snap"
     )
     @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
-    @mock.patch(M_PATH + "LivepatchEntitlement.can_enable", return_value=True)
-    def test_enable_bails_if_snap_cmd_exists_but_snapd_pkg_not_installed(
-        self, m_can_enable, m_app_status, m_which, m_subp, capsys, entitlement
+    @mock.patch(
+        M_PATH + "LivepatchEntitlement.can_enable", return_value=(True, None)
+    )
+    def test_enable_fails_if_snap_cmd_exists_but_snapd_pkg_not_installed(
+        self,
+        m_can_enable,
+        m_app_status,
+        m_which,
+        m_subp,
+        m_livepatch_proxy,
+        m_snap_proxy,
+        m_validate_proxy,
+        capsys,
+        entitlement,
     ):
         """Install canonical-livepatch snap when not present on the system."""
         m_app_status.return_value = status.ApplicationStatus.ENABLED, "enabled"
@@ -548,45 +752,285 @@ class TestLivepatchEntitlementEnable:
             " cannot enable {}".format(entitlement.title)
         )
         assert expected_msg == excinfo.value.msg
+        assert m_validate_proxy.call_count == 0
+        assert m_snap_proxy.call_count == 0
+        assert m_livepatch_proxy.call_count == 0
 
+    @mock.patch("uaclient.apt.get_installed_packages", return_value=["snapd"])
+    @mock.patch("uaclient.util.get_platform_info")
     @mock.patch("uaclient.util.subp")
-    @mock.patch("uaclient.util.which", return_value="/found/livepatch")
+    @mock.patch("uaclient.util.which", side_effect=[True, True])
     @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
-    @mock.patch(M_PATH + "LivepatchEntitlement.can_enable", return_value=True)
+    @mock.patch(
+        M_PATH + "LivepatchEntitlement.can_enable", return_value=(True, None)
+    )
     def test_enable_does_not_install_livepatch_snap_when_present(
-        self, m_can_enable, m_app_status, m_which, m_subp, capsys, entitlement
+        self,
+        m_can_enable,
+        m_app_status,
+        m_which,
+        m_subp,
+        _m_get_platform_info,
+        _m_get_installed_packages,
+        m_livepatch_proxy,
+        m_snap_proxy,
+        m_validate_proxy,
+        capsys,
+        entitlement,
     ):
         """Do not attempt to install livepatch snap when it is present."""
         application_status = status.ApplicationStatus.ENABLED
         m_app_status.return_value = application_status, "enabled"
         assert entitlement.enable()
-        assert self.mocks_config == m_subp.call_args_list
+        subp_calls = [
+            mock.call(
+                [SNAP_CMD, "wait", "system", "seed.loaded"], capture=True
+            ),
+            mock.call(
+                [
+                    LIVEPATCH_CMD,
+                    "config",
+                    "remote-server=https://alt.livepatch.com",
+                ],
+                capture=True,
+            ),
+            mock.call([LIVEPATCH_CMD, "disable"]),
+            mock.call(
+                [LIVEPATCH_CMD, "enable", "livepatch-token"], capture=True
+            ),
+        ]
+        assert subp_calls == m_subp.call_args_list
         assert ("Canonical livepatch enabled.\n", "") == capsys.readouterr()
+        assert m_validate_proxy.call_count == 2
+        assert m_snap_proxy.call_count == 1
+        assert m_livepatch_proxy.call_count == 1
 
+    @mock.patch("uaclient.apt.get_installed_packages", return_value=["snapd"])
+    @mock.patch("uaclient.util.get_platform_info")
     @mock.patch("uaclient.util.subp")
-    @mock.patch("uaclient.util.which", return_value="/found/livepatch")
+    @mock.patch("uaclient.util.which", side_effect=[True, True])
     @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
-    @mock.patch(M_PATH + "LivepatchEntitlement.can_enable", return_value=True)
+    @mock.patch(
+        M_PATH + "LivepatchEntitlement.can_enable", return_value=(True, None)
+    )
     def test_enable_does_not_disable_inactive_livepatch_snap_when_present(
-        self, m_can_enable, m_app_status, m_which, m_subp, capsys, entitlement
+        self,
+        m_can_enable,
+        m_app_status,
+        m_which,
+        m_subp,
+        _m_get_platform_info,
+        _m_get_installed_packages,
+        m_livepatch_proxy,
+        m_snap_proxy,
+        m_validate_proxy,
+        capsys,
+        entitlement,
     ):
         """Do not attempt to disable livepatch snap when it is inactive."""
 
         m_app_status.return_value = status.ApplicationStatus.DISABLED, "nope"
         assert entitlement.enable()
         subp_no_livepatch_disable = [
+            mock.call(
+                [SNAP_CMD, "wait", "system", "seed.loaded"], capture=True
+            ),
             mock.call(
                 [
-                    "/snap/bin/canonical-livepatch",
+                    LIVEPATCH_CMD,
                     "config",
                     "remote-server=https://alt.livepatch.com",
                 ],
                 capture=True,
             ),
             mock.call(
-                ["/snap/bin/canonical-livepatch", "enable", "TOKEN"],
-                capture=True,
+                [LIVEPATCH_CMD, "enable", "livepatch-token"], capture=True
             ),
         ]
         assert subp_no_livepatch_disable == m_subp.call_args_list
         assert ("Canonical livepatch enabled.\n", "") == capsys.readouterr()
+        assert m_validate_proxy.call_count == 2
+        assert m_snap_proxy.call_count == 1
+        assert m_livepatch_proxy.call_count == 1
+
+    @pytest.mark.parametrize(
+        "cls_name, cls_title", (("FIPSEntitlement", "FIPS"),)
+    )
+    @mock.patch("uaclient.util.handle_message_operations")
+    @mock.patch("uaclient.util.is_container", return_value=False)
+    def test_enable_fails_when_blocking_service_is_enabled(
+        self,
+        m_is_container,
+        m_handle_message_op,
+        m_livepatch_proxy,
+        m_snap_proxy,
+        m_validate_proxy,
+        cls_name,
+        cls_title,
+        entitlement,
+    ):
+        m_handle_message_op.return_value = True
+
+        with mock.patch(M_LIVEPATCH_STATUS, return_value=DISABLED_APP_STATUS):
+            with mock.patch(
+                "uaclient.entitlements.fips.{}.application_status".format(
+                    cls_name
+                )
+            ) as m_fips:
+                m_fips.return_value = (status.ApplicationStatus.ENABLED, "")
+                result, reason = entitlement.enable()
+                assert not result
+
+                msg = "Cannot enable Livepatch when {} is enabled.".format(
+                    cls_title
+                )
+                assert msg.strip() == reason.message.msg.strip()
+
+        assert m_validate_proxy.call_count == 0
+        assert m_snap_proxy.call_count == 0
+        assert m_livepatch_proxy.call_count == 0
+
+    @pytest.mark.parametrize("caplog_text", [logging.WARN], indirect=True)
+    @mock.patch("uaclient.util.which")
+    @mock.patch("uaclient.apt.get_installed_packages")
+    @mock.patch("uaclient.util.subp")
+    def test_enable_alerts_user_that_snapd_does_not_wait_command(
+        self,
+        m_subp,
+        m_installed_pkgs,
+        m_which,
+        m_livepatch_proxy,
+        m_snap_proxy,
+        m_validate_proxy,
+        entitlement,
+        capsys,
+        caplog_text,
+    ):
+        m_which.side_effect = [True, False]
+        m_installed_pkgs.return_value = ["snapd"]
+        stderr_msg = (
+            "error: Unknown command `wait'. Please specify one command of: "
+            "abort, ack, buy, change, changes, connect, create-user, disable,"
+            " disconnect, download, enable, find, help, install, interfaces, "
+            "known, list, login, logout, refresh, remove, run or try"
+        )
+
+        m_subp.side_effect = [
+            exceptions.ProcessExecutionError(
+                cmd="snapd wait system seed.loaded",
+                exit_code=-1,
+                stdout="",
+                stderr=stderr_msg,
+            ),
+            True,
+        ]
+
+        fake_stdout = io.StringIO()
+
+        with mock.patch.object(entitlement, "can_enable") as m_can_enable:
+            m_can_enable.return_value = (True, None)
+            with mock.patch.object(
+                entitlement, "setup_livepatch_config"
+            ) as m_setup_livepatch:
+                with contextlib.redirect_stdout(fake_stdout):
+                    entitlement.enable()
+
+                assert 1 == m_can_enable.call_count
+                assert 1 == m_setup_livepatch.call_count
+
+        assert (
+            "Installing canonical-livepatch snap"
+            in fake_stdout.getvalue().strip()
+        )
+
+        for msg in messages.SNAPD_DOES_NOT_HAVE_WAIT_CMD.split("\n"):
+            assert msg in caplog_text()
+
+        assert m_validate_proxy.call_count == 2
+        assert m_snap_proxy.call_count == 1
+        assert m_livepatch_proxy.call_count == 1
+
+    @mock.patch("uaclient.util.which")
+    @mock.patch("uaclient.apt.get_installed_packages")
+    @mock.patch("uaclient.util.subp")
+    def test_enable_raise_exception_for_unexpected_error_on_snapd_wait(
+        self,
+        m_subp,
+        m_installed_pkgs,
+        m_which,
+        m_livepatch_proxy,
+        m_snap_proxy,
+        m_validate_proxy,
+        entitlement,
+    ):
+        m_which.side_effect = [False, True]
+        m_installed_pkgs.return_value = ["snapd"]
+        stderr_msg = "test error"
+
+        m_subp.side_effect = exceptions.ProcessExecutionError(
+            cmd="snapd wait system seed.loaded",
+            exit_code=-1,
+            stdout="",
+            stderr=stderr_msg,
+        )
+
+        with mock.patch.object(entitlement, "can_enable") as m_can_enable:
+            m_can_enable.return_value = (True, None)
+            with mock.patch.object(
+                entitlement, "setup_livepatch_config"
+            ) as m_setup_livepatch:
+                with pytest.raises(
+                    exceptions.ProcessExecutionError
+                ) as excinfo:
+                    entitlement.enable()
+
+            assert 1 == m_can_enable.call_count
+            assert 0 == m_setup_livepatch.call_count
+
+        expected_msg = "test error"
+        assert expected_msg in str(excinfo)
+        assert m_validate_proxy.call_count == 0
+        assert m_snap_proxy.call_count == 0
+        assert m_livepatch_proxy.call_count == 0
+
+
+class TestLivepatchApplicationStatus:
+    @pytest.mark.parametrize("which_result", ((True), (False)))
+    @pytest.mark.parametrize("subp_raise_exception", ((True), (False)))
+    @mock.patch("uaclient.util.which")
+    @mock.patch("uaclient.util.subp")
+    def test_application_status(
+        self, m_subp, m_which, subp_raise_exception, which_result, entitlement
+    ):
+        m_which.return_value = which_result
+
+        if subp_raise_exception:
+            m_subp.side_effect = exceptions.ProcessExecutionError("error msg")
+
+        status, details = entitlement.application_status()
+
+        if not which_result:
+            assert status == ApplicationStatus.DISABLED
+            assert "canonical-livepatch snap is not installed." in details.msg
+        elif subp_raise_exception:
+            assert status == ApplicationStatus.DISABLED
+            assert "error msg" in details.msg
+        else:
+            assert status == ApplicationStatus.ENABLED
+            assert details is None
+
+    @mock.patch("time.sleep")
+    @mock.patch("uaclient.util.which", return_value=True)
+    def test_status_command_retry_on_application_status(
+        self, m_which, m_sleep, entitlement
+    ):
+        from uaclient import util
+
+        with mock.patch.object(util, "_subp") as m_subp:
+            m_subp.side_effect = exceptions.ProcessExecutionError("error msg")
+            status, details = entitlement.application_status()
+
+            assert m_subp.call_count == 3
+            assert m_sleep.call_count == 2
+            assert status == ApplicationStatus.DISABLED
+            assert "error msg" in details.msg
