@@ -1,32 +1,39 @@
 """Tests related to uaclient.apt module."""
 
 import glob
-import mock
 import os
 import stat
 import subprocess
 from textwrap import dedent
 
+import mock
 import pytest
 
+from uaclient import apt, exceptions, messages, util
 from uaclient.apt import (
-    APT_KEYS_DIR,
     APT_AUTH_COMMENT,
+    APT_CONFIG_PROXY_HTTP,
+    APT_CONFIG_PROXY_HTTPS,
+    APT_KEYS_DIR,
+    APT_PROXY_CONF_FILE,
+    APT_RETRIES,
     KEYRINGS_DIR,
     add_apt_auth_conf_entry,
     add_auth_apt_repo,
     add_ppa_pinning,
-    clean_apt_sources,
+    assert_valid_apt_credentials,
+    clean_apt_files,
     find_apt_list_files,
     get_installed_packages,
     remove_apt_list_files,
     remove_auth_apt_repo,
     remove_repo_from_apt_auth_file,
-    assert_valid_apt_credentials,
+    run_apt_update_command,
+    setup_apt_proxy,
 )
-from uaclient import apt, exceptions, util
-from uaclient.entitlements.tests.test_base import ConcreteTestEntitlement
-
+from uaclient.entitlements.base import UAEntitlement
+from uaclient.entitlements.repo import RepoEntitlement
+from uaclient.entitlements.tests.test_repo import RepoTestEntitlement
 
 POST_INSTALL_APT_CACHE_NO_UPDATES = """
 -32768 https://esm.ubuntu.com/ubuntu/ {0}-updates/main amd64 Packages
@@ -145,7 +152,8 @@ class TestValidAptCredentials:
                 "http://user:pwd@fakerepo/ubuntu/pool/",
                 expected_path,
             ],
-            timeout=20,
+            timeout=60,
+            retry_sleeps=APT_RETRIES,
         )
         assert [apt_helper_call] == m_subp.call_args_list
 
@@ -191,7 +199,7 @@ class TestValidAptCredentials:
             "/does/not/exist"
         )
         # Failure apt-helper response
-        m_subp.side_effect = util.ProcessExecutionError(
+        m_subp.side_effect = exceptions.ProcessExecutionError(
             cmd="apt-helper ",
             exit_code=exit_code,
             stdout="Err:1...",
@@ -216,7 +224,8 @@ class TestValidAptCredentials:
                 "http://user:pwd@fakerepo/ubuntu/pool/",
                 expected_path,
             ],
-            timeout=20,
+            timeout=60,
+            retry_sleeps=APT_RETRIES,
         )
         assert [apt_helper_call] == m_subp.call_args_list
 
@@ -259,6 +268,7 @@ class TestValidAptCredentials:
                 expected_path,
             ],
             timeout=apt.APT_HELPER_TIMEOUT,
+            retry_sleeps=APT_RETRIES,
         )
         assert [apt_helper_call] == m_subp.call_args_list
 
@@ -541,31 +551,38 @@ class TestAddAptAuthConfEntry:
         assert expected_content == util.load_file(auth_file)
 
 
-class TestCleanAptSources:
-    @pytest.fixture
-    def mock_apt_entitlement(self, tmpdir):
+class TestCleanAptFiles:
+    @pytest.fixture(params=[RepoEntitlement, UAEntitlement])
+    def mock_apt_entitlement(self, request, tmpdir):
         # Set up our tmpdir with some fake list files
         entitlement_name = "test_ent"
-        file_tmpl = tmpdir.join("{name}-{series}").strpath
+        repo_tmpl = tmpdir.join("source-{name}").strpath
+        pref_tmpl = tmpdir.join("pref-{name}").strpath
+
+        class TestRepo(request.param):
+            name = entitlement_name
+            repo_list_file_tmpl = repo_tmpl
+            repo_pref_file_tmpl = pref_tmpl
+            is_repo = request.param == RepoEntitlement
+
         for series in ["acidic", "base"]:
-            file_name = file_tmpl.format(name=entitlement_name, series=series)
-            with open(file_name, "w") as f:
+            source_name = repo_tmpl.format(name=entitlement_name)
+            pref_name = pref_tmpl.format(name=entitlement_name)
+
+            with open(source_name, "w") as f:
                 f.write("")
 
-        m_entitlement = mock.Mock(spec=ConcreteTestEntitlement)
-        m_entitlement.configure_mock(
-            name=entitlement_name,
-            repo_url="some url",
-            repo_list_file_tmpl=file_tmpl,
-        )
-        return m_entitlement
+            with open(pref_name, "w") as f:
+                f.write("")
+
+        return TestRepo
 
     @mock.patch("uaclient.apt.os.unlink")
     def test_no_removals_for_no_repo_entitlements(self, m_os_unlink):
         m_entitlements = mock.Mock()
-        m_entitlements.ENTITLEMENT_CLASSES = [ConcreteTestEntitlement]
+        m_entitlements.ENTITLEMENT_CLASSES = [RepoTestEntitlement]
 
-        clean_apt_sources(_entitlements=m_entitlements)
+        clean_apt_files(_entitlements=m_entitlements)
 
         assert 0 == m_os_unlink.call_count
 
@@ -573,9 +590,14 @@ class TestCleanAptSources:
         m_entitlements = mock.Mock()
         m_entitlements.ENTITLEMENT_CLASSES = [mock_apt_entitlement]
 
-        clean_apt_sources(_entitlements=m_entitlements)
+        clean_apt_files(_entitlements=m_entitlements)
 
-        assert [] == tmpdir.listdir()
+        if mock_apt_entitlement.is_repo:
+            assert [] == tmpdir.listdir()
+        else:
+            assert sorted(
+                [tmpdir.join("source-test_ent"), tmpdir.join("pref-test_ent")]
+            ) == sorted(tmpdir.listdir())
 
     def test_other_files_not_removed(self, mock_apt_entitlement, tmpdir):
         other_filename = "other_file-acidic"
@@ -584,9 +606,18 @@ class TestCleanAptSources:
         m_entitlements = mock.Mock()
         m_entitlements.ENTITLEMENT_CLASSES = [mock_apt_entitlement]
 
-        clean_apt_sources(_entitlements=m_entitlements)
+        clean_apt_files(_entitlements=m_entitlements)
 
-        assert [tmpdir.join(other_filename)] == tmpdir.listdir()
+        if mock_apt_entitlement.is_repo:
+            assert [tmpdir.join(other_filename)] == tmpdir.listdir()
+        else:
+            assert sorted(
+                [
+                    tmpdir.join("source-test_ent"),
+                    tmpdir.join("pref-test_ent"),
+                    tmpdir.join(other_filename),
+                ]
+            ) == sorted(tmpdir.listdir())
 
 
 @pytest.fixture(params=(mock.sentinel.default, None, "some_string"))
@@ -820,3 +851,142 @@ class TestGetInstalledPackages:
     @mock.patch("uaclient.apt.util.subp", return_value=("a\nb", ""))
     def test_assert_missing_eof_newline_works(self, m_subp):
         assert ["a", "b"] == get_installed_packages()
+
+
+class TestRunAptCommand:
+    @pytest.mark.parametrize(
+        "error_list, output_list",
+        (
+            (
+                [
+                    "E: The repository 't1 404' does not have a Release.",
+                    "W: Failed to fetch t1/dists/ 404 Not found.",
+                    "E: The repository 't2 404' does not have a Release.\n",
+                ],
+                (
+                    "APT update failed.",
+                    (
+                        "APT update failed to read APT config "
+                        "for the following URLs:"
+                    ),
+                    "- t1",
+                    "- t2",
+                ),
+            ),
+            (
+                ["E: The repository 't1 404' does not have a Release file."],
+                (
+                    "APT update failed.",
+                    (
+                        "APT update failed to read APT config "
+                        "for the following URL:"
+                    ),
+                    "- t1",
+                ),
+            ),
+            (
+                [
+                    "W: Failed to fetch t1/dists Not Found [IP: 127.0.0.1]\n",
+                    "E: Some index files failed to download.\n",
+                ],
+                (
+                    "APT update failed.",
+                    (
+                        "APT update failed to read APT config "
+                        "for the following URL:"
+                    ),
+                    "- t1",
+                ),
+            ),
+        ),
+    )
+    @mock.patch("uaclient.apt.util.subp")
+    def test_run_apt_command_with_invalid_repositories(
+        self, m_subp, error_list, output_list
+    ):
+        error_msg = "\n".join(error_list)
+
+        m_subp.side_effect = exceptions.ProcessExecutionError(
+            cmd="apt update", stderr=error_msg
+        )
+
+        with pytest.raises(exceptions.UserFacingError) as excinfo:
+            run_apt_update_command()
+
+        expected_message = "\n".join(output_list) + "."
+        assert expected_message == excinfo.value.msg
+
+
+class TestAptProxyConfig:
+    @pytest.mark.parametrize(
+        "kwargs, expected_remove_calls, expected_write_calls, expected_out",
+        [
+            ({}, [mock.call(APT_PROXY_CONF_FILE)], [], ""),
+            (
+                {"http_proxy": "mock_http_proxy"},
+                [],
+                [
+                    mock.call(
+                        APT_PROXY_CONF_FILE,
+                        messages.APT_PROXY_CONFIG_HEADER
+                        + APT_CONFIG_PROXY_HTTP.format(
+                            proxy_url="mock_http_proxy"
+                        ),
+                    )
+                ],
+                messages.SETTING_SERVICE_PROXY.format(service="APT"),
+            ),
+            (
+                {"https_proxy": "mock_https_proxy"},
+                [],
+                [
+                    mock.call(
+                        APT_PROXY_CONF_FILE,
+                        messages.APT_PROXY_CONFIG_HEADER
+                        + APT_CONFIG_PROXY_HTTPS.format(
+                            proxy_url="mock_https_proxy"
+                        ),
+                    )
+                ],
+                messages.SETTING_SERVICE_PROXY.format(service="APT"),
+            ),
+            (
+                {
+                    "http_proxy": "mock_http_proxy",
+                    "https_proxy": "mock_https_proxy",
+                },
+                [],
+                [
+                    mock.call(
+                        APT_PROXY_CONF_FILE,
+                        messages.APT_PROXY_CONFIG_HEADER
+                        + APT_CONFIG_PROXY_HTTP.format(
+                            proxy_url="mock_http_proxy"
+                        )
+                        + APT_CONFIG_PROXY_HTTPS.format(
+                            proxy_url="mock_https_proxy"
+                        ),
+                    )
+                ],
+                messages.SETTING_SERVICE_PROXY.format(service="APT"),
+            ),
+        ],
+    )
+    @mock.patch("uaclient.util.write_file")
+    @mock.patch("uaclient.util.remove_file")
+    def test_setup_apt_proxy_config(
+        self,
+        m_util_remove_file,
+        m_util_write_file,
+        kwargs,
+        expected_remove_calls,
+        expected_write_calls,
+        expected_out,
+        capsys,
+    ):
+        setup_apt_proxy(**kwargs)
+        assert expected_remove_calls == m_util_remove_file.call_args_list
+        assert expected_write_calls == m_util_write_file.call_args_list
+        out, err = capsys.readouterr()
+        assert expected_out == out.strip()
+        assert "" == err
